# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import iocage.lib.helpers


class ZFSBasejailStorage:
    def prepare(self):
        self._delete_clone_target_datasets()

    def apply(self, release=None):

        if release is None:
            release = self.jail.cloned_release

        return ZFSBasejailStorage.clone(self, release)

    def setup(self, release):
        iocage.lib.StandaloneJailStorage.StandaloneJailStorage.setup(
            self, release)

    def clone(self, release):

        current_basejail_type = self.jail.config["basejail_type"]
        if not current_basejail_type == "zfs":
            raise iocage.lib.errors.InvalidJailConfigValue(
                property_name="basejail_type",
                reason="Expected ZFS, but saw {current_basejail_type}",
                logger=self.logger
            )

        ZFSBasejailStorage._create_mountpoints(self)

        basedirs = iocage.lib.helpers.get_basedir_list(
            distribution_name=self.jail.host.distribution.name
        )

        for basedir in basedirs:
            source_dataset_name = f"{release.base_dataset.name}/{basedir}"
            target_dataset_name = f"{self.jail.root_dataset_name}/{basedir}"
            self.clone_zfs_dataset(source_dataset_name, target_dataset_name)

    def _delete_clone_target_datasets(self, root=None):

        if root is None:
            root = list(self.jail.root_dataset.children)

        for child in root:
            root_dataset_prefix = f"{self.jail.root_dataset_name}/"
            relative_name = child.name.replace(root_dataset_prefix, "")
            basedirs = iocage.lib.helpers.get_basedir_list()

            if relative_name in basedirs:

                # Unmount if mounted
                try:
                    current_mountpoint = child.mountpoint
                    if current_mountpoint:
                        child.umount()
                        self.logger.verbose(
                            f"Clone target {current_mountpoint} unmounted"
                        )
                except:
                    pass

                # Delete existing snapshots
                for snapshot in child.snapshots:
                    try:
                        snapshot.delete()
                        self.logger.verbose(
                            f"Snapshot {current_mountpoint} deleted"
                        )
                    except:
                        pass

                child.delete()

            else:
                self._delete_clone_target_datasets(list(child.children))

    def _create_mountpoints(self):
        for basedir in ["dev", "etc"]:
            self.create_jail_mountpoint(basedir)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import shutil

import iocage.lib.helpers
import iocage.lib.Config.Jail


class ResolverProp(list):

    config: 'iocage.lib.Config.Jail.JailConfig.JailConfig'  # type: ignore
    property_name: str = "resolver"

    def __init__(
        self,
        config,  # type: ignore
        logger: 'iocage.lib.Logger.Logger'=None,
        **kwargs
    ) -> None:

        list.__init__(self, [])
        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.config = config
        self.config.attach_special_property(
            name="resolver",
            special_property=self
        )

    @property
    def conf_file_path(self):
        return "/etc/resolv.conf"

    @property
    def method(self):
        return self._get_method(self.value)

    def _get_method(self, value: str) -> str:
        if value == "/etc/resolv.conf":
            return "copy"

        elif value == "/dev/null":
            return "skip"

        else:
            return "manual"

    @property
    def value(self):
        return self.config.data["resolver"]

    def apply(self, jail):

        self.logger.verbose(
            f"Configuring nameserver for Jail '{jail.humanreadable_name}'"
        )

        remote_path = f"{jail.root_path}/{self.conf_file_path}"

        if self.method == "copy":
            shutil.copy(self.conf_file_path, remote_path)
            self.logger.verbose("resolv.conf copied from host")

        elif self.method == "manual":
            with open(remote_path, "w") as f:
                f.write("\n".join(self))
                f.close()
            self.logger.verbose("resolv.conf written manually")

        else:
            self.logger.verbose("resolv.conf not touched")

    def set(self, value=None, notify=True):

        self.clear()
        method = self._get_method(value)
        if method == "manual":
            if isinstance(value, str):
                self += value.split(";")
            elif isinstance(value, list):
                self += value
            else:
                raise TypeError("value can be list or string")
        else:
            self.append(value, notify=False)

        self.__notify(notify)

    def append(self, value, notify=True):
        list.append(self, value)
        self.__notify(notify)

    def __setitem__(self, key, value, notify=True):
        list.__setitem__(self, key, value)
        self.__notify(notify)

    def __str__(self):
        out = ";".join(list(self))
        return out

    def __notify(self, notify=True):

        if not notify:
            return

        try:
            self.config.update_special_property("resolver")
        except:
            raise



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import os.path
import re

import iocage.lib.errors
import iocage.lib.helpers


class DevfsRulesFilter:

    def __init__(self, source, active_filters=[]):
        self.source = source
        self.active_filters = active_filters

    def with_rule(self, rule):
        return DevfsRulesFilter(filter(
            lambda x: x.has_rule(rule),
            self.source
        ), self.active_filters + [rule])

    def with_include(self, rule_name):
        rule = f"add include ${rule_name}"
        return self.with_rule(rule)

    def create_ruleset(self, ruleset_name, ruleset_number):
        """
        Create a ruleset from the provided filters

        Args:

            ruleset_name (string):
                Name of the rule that get's created

            ruleset_number (int):
                Number of the rule that get's created
        """

        ruleset = DevfsRuleset(ruleset_name, ruleset_number)
        for line in self.active_filters:
            ruleset.append(line)
        return ruleset


class DevfsRuleset(list):
    """
    Representation of a devfs ruleset in the devfs.rules file

    DevfsRuleset instances behave like standard lists and can store strings
    that multiple lines (string)
    """

    PATTERN = re.compile(r"""^\[(?P<name>[a-z](?:[a-z0-9\-_]*[a-z0-9])?)=
                (?P<number>[0-9]+)\]\s*(?:\#\s*(?P<comment>.*))?$""", re.X)

    def __init__(self, value=None, number=None, comment=None):
        """
        Initialize DevfsRuleset

        Args:

            value (string): (optional)
                If specified in combination with a number, this parameter is
                interpreted as the ruleset name. Otherwise it is parsed with
                the expectation to find a [<name>=<number>] ruleset definition.

                When value is not specified or None, DevfsRuleset is assumed
                to be new or unspecified (cannot be exported before name and
                number were assigned at a later time)

            number (int): (optional)
                The number of the ruleset. Must be specified to export the
                ruleset, but is like the name not required to compare the
                ruleset with another
        """

        # when only one argument is passed, it's a line that need to be parsed
        if value is None and number is None:
            # name and number will be assigned later
            name = None
        elif number is None:
            name, number, comment = self._parse_line(value)
        else:
            name = int(value)

        self.name = name
        self.number = number
        self.comment = comment
        list.__init__(self)

    def has_rule(self, rule):
        """
        Returns true if the rule is part of the current ruleset

        Args:

            rule (string):
                The rule string to be compared with current rules of the
                ruleset instance
        """
        return rule in self

    def append(self, rule):
        if rule not in self:
            list.append(self, rule)

    def clone(self, source_ruleset):
        """
        Clones the rules from another ruleset

        Args:

            source_ruleset (iocage.lib.DevfsRules.DevfsRuleset):
                Ruleset to copy all rules from
        """
        for rule in source_ruleset:
            self.append(rule)

    def _parse_line(self, line):

        # marks beginning of a new ruleset
        ruleset_match = re.search(DevfsRuleset.PATTERN, line)
        if ruleset_match is not None:
            name = str(ruleset_match.group("name"))
            number = int(ruleset_match.group("number"))
            comment = ruleset_match.group("comment")
            return name, number, comment

        raise SyntaxError("DevfsRuleset line parsing failed")

    def __str__(self):
        ruleset_line = f"[{self.name}={self.number}]"
        if self.comment is not None:
            ruleset_line += f" # {self.comment}"
        output = [ruleset_line] + [str(x) for x in self]
        return "\n".join(output) + "\n"


class DevfsRules(list):
    """
    Abstraction for the hosts /etc/devfs.rules

    Read and edit devfs rules in a programmatic way.
    Restarts devfs service after applying changes.
    """

    def __init__(self, rules_file="/etc/devfs.rules", logger=None):
        """
        Initializes a DevfsRules manager for devfs.rules files

        Args:

            rules_file (string): (default=/etc/devfs.rules)
                Path of the devfs.rules file

            logger (iocage.Logger): (optional)
                Instance of the logger that is passed to occuring errors
        """

        self.logger = logger

        # index rulesets to find duplicated and provide easy access
        self._ruleset_number_index = {}
        self._ruleset_name_index = {}

        # remember all lines that were loaded from defaults (system)
        self._system_rule_lines = []

        list.__init__(self)

        # will automatically read from file - needs to be the last item
        self.rules_file = rules_file

    def append(self, ruleset, is_system_rule=False):
        """
        Add a DevfsRuleset to the list

        The rulesets added become indexed, so that lookups and duplication
        checks are easy and fast

        Args:

            ruleset (iocage.lib.DevfsRules.DevfsRuleset|string):
                The ruleset that gets added if it is not already in the list
        """

        next_line_index = len(self)

        if ruleset is None or isinstance(ruleset, str):
            list.append(self, ruleset)
            if is_system_rule is True:
                self._system_rule_lines.append(next_line_index)
            return ruleset

        if ruleset.name in self._ruleset_name_index.keys():
            raise iocage.lib.errors.DuplicateDevfsRuleset(
                reason=f"Ruleset named '{ruleset.name}' already present",
                devfs_rules_file=self.rules_file,
                logger=self.logger
            )

        if ruleset.number in self._ruleset_number_index.keys():
            raise iocage.lib.errors.DuplicateDevfsRuleset(
                reason=f"Ruleset number '{ruleset.number}' already present",
                devfs_rules_file=self.rules_file,
                logger=self.logger
            )

        # build indexes
        self._ruleset_number_index[ruleset.number] = next_line_index
        self._ruleset_name_index[ruleset.name] = next_line_index
        if is_system_rule is True:
            self._system_rule_lines.append(next_line_index)

        list.append(self, ruleset)
        return ruleset

    def new_ruleset(self, ruleset):
        """
        Append a new ruleset

        Similar to append(), but automatically assigns a new number

        Args:

            ruleset (iocage.lib.DevfsRules.DevfsRuleset):
                The new devfs ruleset that is going to be added

        Returns:

            int: The devfs ruleset number of the created ruleset
        """

        ruleset.number = self.next_number

        if ruleset.name is None:
            ruleset.name = f"iocage_auto_{ruleset.number}"

        self.append(ruleset)
        return ruleset.number

    def find_by_name(self, rule_name):
        return self._find_by_index(rule_name, self._ruleset_name_index)

    def find_by_number(self, rule_number):
        return self._find_by_index(rule_number, self._ruleset_number_index)

    def _find_by_index(self, rule_name, index):
        return self[index[rule_name]]

    @property
    def default_rules_file(self):
        return "/etc/defaults/devfs.rules"

    @property
    def rules_file(self):
        """
        Path of the devfs.rules file
        """
        return self._rules_file

    @rules_file.setter
    def rules_file(self, devfs_rules_path):
        """
        When setting a new devfs.rules source, it is read automatically
        """
        self._rules_file = devfs_rules_path
        try:
            self.read_rules()
        except FileNotFoundError:
            pass

    @property
    def next_number(self):
        """
        The next highest ruleset number that is available

        This counting includes the systems default devfs rulesets
        """
        return len(self._ruleset_name_index.keys()) + 1

    def read_rules(self):
        """
        Read existing devfs.rules file

        Existing devfs rules get reset and read from the rules_file
        """

        if self.logger:
            self.logger.debug(f"Reading devfs.rules from {self.rules_file}")

        self.clear()
        self._read_rules_file(self.default_rules_file, system=True)
        self._read_rules_file(self.rules_file)

    def _read_rules_file(self, file, system=False):

        f = open(file, "r")

        current_ruleset = None

        for line in f.readlines():

            line = line.strip().rstrip("\n")

            # add comments and empty lines as string
            if line.startswith("#") or (line == ""):
                self.append(line, is_system_rule=system)
                continue

            try:
                current_ruleset = DevfsRuleset(line)
                self.append(current_ruleset, is_system_rule=system)
                continue
            except SyntaxError:
                pass

            # the first item must be a ruleset
            if current_ruleset is None:
                raise iocage.lib.errors.InvalidDevfsRulesSyntax(
                    devfs_rules_file=self.rules_file,
                    reason="Rules must follow a ruleset declaration",
                    logger=self.logger
                )

            current_ruleset.append(line)

        f.close()

    def save(self):
        """
        Apply changes to the devfs.rules file

        Automatically restarts devfs service when the file was changed
        """

        content_before = None

        if os.path.isfile(self.rules_file):
            f = open(self.rules_file, "r+")
            content_before = f.read()
            f.seek(0)
        else:
            f = open(self.rules_file, "w")

        new_content = self.__str__()

        if content_before == new_content:
            if self.logger is not None:
                self.logger.verbose(
                    f"devfs.rules file {self.rules_file} unchanged"
                )
        else:
            if self.logger is not None:
                self.logger.verbose(
                    f"Writing devfs.rules to {self.rules_file}"
                )
                self.logger.spam(new_content, indent=1)

            f.write(new_content)
            f.truncate()
            self._restart_devfs_service()

        f.close()

    def _restart_devfs_service(self):
        """
        Restart devfs service after changing devfs.rules
        """
        if self.logger is not None:
            self.logger.debug("Restarting devfs service")
        iocage.lib.helpers.exec(["service", "devfs", "restart"])

    def __str__(self):
        """
        Output the devfs.rules content as string
        """

        out_lines = []
        for i, line in enumerate(self):
            if i not in self._system_rule_lines:
                out_lines.append(str(line))

        return "\n".join(out_lines)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import os
import sys

import iocage.lib.errors

from typing import List


class LogEntry:

    def __init__(self, message, level, indent=0, logger=None, **kwargs):
        self.message = message
        self.level = level
        self.indent = indent
        self.logger = logger

        for key in kwargs.keys():
            object.__setattr__(self, key, kwargs[key])

    def edit(self, message=None, indent=None):

        if self.logger is None:
            raise iocage.lib.errors.CannotRedrawLine(
                reason="No logger available"
            )

        if message is not None:
            self.message = message

        if indent is not None:
            self.indent = indent

        self.logger.redraw(self)

    def __len__(self):
        return len(self.message.split("\n"))


class Logger:

    COLORS = (
        "black",
        "red",
        "green",
        "yellow",
        "blue",
        "margenta",
        "cyan",
        "white",
    )

    RESET_SEQ = "\033[0m"
    BOLD_SEQ = "\033[1m"
    LINE_UP_SEQ = "\033[F"

    LOG_LEVEL_SETTINGS = {
        "screen": {"color": None},
        "info": {"color": None},
        "notice": {"color": "magenta"},
        "verbose": {"color": "blue"},
        "spam": {"color": "green"},
        "critical": {"color": "red", "bold": True},
        "error": {"color": "red"},
        "debug": {"color": "green"},
        "warn": {"color": "yellow"}
    }

    LOG_LEVELS = (
        "critical",
        "error",
        "warn",
        "info",
        "notice",
        "verbose",
        "debug",
        "spam",
        "screen"
    )

    INDENT_PREFIX = "  "

    PRINT_HISTORY: List[str] = []

    def __init__(self, print_level=None, log_directory="/var/log/iocage"):
        self._print_level = print_level
        self._set_log_directory(log_directory)

    @property
    def default_print_level(self):
        return "info"

    @property
    def print_level(self):
        if self._print_level is None:
            return self.default_print_level
        else:
            return self._print_level

    @print_level.setter
    def print_level(self, value):
        self._print_level = value

    def _set_log_directory(self, log_directory):
        self.log_directory = os.path.abspath(log_directory)
        if not os.path.isdir(log_directory):
            self._create_log_directory()
        self.log(f"Log directory set to '{log_directory}'", level="spam")

    def log(self, *args, **kwargs):

        args = list(args)

        if ("message" not in kwargs) and (len(args) > 0):
            kwargs["message"] = args.pop(0)

        if ("level" not in kwargs) and (len(args) > 0):
            kwargs["level"] = args.pop(0)

        if "level" not in kwargs:
            kwargs["level"] = "info"

        log_entry = LogEntry(logger=self, **kwargs)

        if self._should_print_log_entry(log_entry):
            self._print_log_entry(log_entry)
            self.PRINT_HISTORY.append(log_entry)

        return log_entry

    def verbose(self, message, indent=0, **kwargs):
        return self.log(message, level="verbose", indent=indent, **kwargs)

    def error(self, message, indent=0, **kwargs):
        return self.log(message, level="error", indent=indent, **kwargs)

    def warn(self, message, indent=0, **kwargs):
        return self.log(message, level="warn", indent=indent, **kwargs)

    def debug(self, message, indent=0, **kwargs):
        return self.log(message, level="debug", indent=indent, **kwargs)

    def spam(self, message, indent=0, **kwargs):
        return self.log(message, level="spam", indent=indent, **kwargs)

    def screen(self, message, indent=0, **kwargs):
        """
        Screen never gets printed to log files
        """
        return self.log(message, level="screen", indent=indent, **kwargs)

    def redraw(self, log_entry):

        if log_entry not in self.PRINT_HISTORY:
            raise iocage.lib.errors.CannotRedrawLine(
                reason="Log entry not found in history"
            )

        if log_entry.level != "screen":
            raise iocage.lib.errors.CannotRedrawLine(
                reason=(
                    "Log level 'screen' is required to redraw, "
                    f"but got '{self.level}'"
                )
            )

        # calculate the delta of messages printed since
        i = self.PRINT_HISTORY.index(log_entry)
        n = len(self.PRINT_HISTORY)
        delta = sum(map(lambda i: len(self.PRINT_HISTORY[i]), range(i, n)))

        output = "".join([
            "\r",
            f"\033[{delta}F",  # CPL - Cursor Previous Line
            "\r",               # CR - Carriage Return
            self._indent(
                f"{log_entry.message}: {delta}",
                log_entry.indent
            ),
            "\033[K",           # EL - Erase in Line
            "\n" * (delta),
            "\r"
        ])

        sys.stdout.write(output)

    def _should_print_log_entry(self, log_entry):

        if log_entry.level == "screen":
            return True

        if self.print_level is False:
            return False

        print_level = Logger.LOG_LEVELS.index(self.print_level)
        return Logger.LOG_LEVELS.index(log_entry.level) <= print_level

    def _beautify_message(self, message, level, indent=0):
        color = self._get_level_color(level)
        message = self._indent(message, indent)
        message = self._colorize(message, color)
        return message

    def _print(self, message, level, indent=0):
        print(self._beautify_message(message, level, indent))

    def _print_log_entry(self, log_entry):
        return self._print(
            log_entry.message,
            log_entry.level,
            log_entry.indent
        )

    def _indent(self, message, level):
        indent = Logger.INDENT_PREFIX * level
        return "\n".join(map(lambda x: f"{indent}{x}", message.split("\n")))

    def _get_log_file_path(self, level, jail=None):
        return self.log_directory

    def _create_log_directory(self):
        if os.geteuid() != 0:
            raise iocage.lib.errors.MustBeRoot(
                f"create {self.log_directory}")
        os.makedirs(self.log_directory, 0x600)
        self.log(f"Log directory '{self.log_directory}' created", level="info")

    def _get_color_code(self, color_name):
        return Logger.COLORS.index(color_name) + 30

    def _get_level_color(self, log_level):
        try:
            return Logger.LOG_LEVEL_SETTINGS[log_level]["color"]
        except KeyError:
            return "none"

    def _colorize(self, message, color_name=None):
        try:
            color_code = self._get_color_code(color_name)
        except:
            return message

        return f"\033[1;{color_code}m{message}\033[0m"



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing

import iocage.lib.helpers

_ConfigType = 'iocage.lib.Config.Jail.JailConfig.JailConfig'


class BridgeSet(set):

    def __init__(self, config=None):
        self.config = config
        set.__init__(self)

    def add(self, value, notify=True):
        set.add(self, value)
        if notify:
            self.notify()

    def remove(self, value, notify=True):
        set.remove(self, value)
        if notify:
            self._notify()

    def _notify(self):
        try:
            self.config.update_special_property("interfaces")
        except:
            pass


class InterfaceProp(dict):

    config: _ConfigType  # type: ignore
    property_name: str = "interfaces"

    def __init__(
        self,
        config=None,
        **kwargs
    ) -> None:

        dict.__init__(self, {})
        self.config = config

    def set(
        self,
        data: typing.Union[str, typing.Dict[str, BridgeSet]]
    ) -> None:

        self.clear()

        try:
            iocage.lib.helpers.parse_none(data)
            return
        except TypeError:
            pass

        if isinstance(data, dict):
            dict.__init__(self, data)
            return

        nic_pairs = data.replace(",", " ").split(" ")
        for nic_pair in nic_pairs:
            jail_if, bridge_if = nic_pair.split(":", maxsplit=1)
            self.add(jail_if, bridge_if, notify=False)

        self.__notify()

    def add(self, jail_if, bridges=None, notify=True):

        if bridges is None or bridges == [] or bridges == "":
            return

        if isinstance(bridges, str):
            bridges = [bridges]

        try:
            prop = dict.__getitem__(self, jail_if)
        except:
            prop = self.__empty_prop(jail_if)

        for bridge_if in bridges:
            prop.add(bridge_if, notify=False)

        if notify:
            self.__notify()

    def __setitem__(self, key, values):

        try:
            dict.__delitem__(self, key)
        except:
            pass

        self.add(key, values)

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self.__notify()

    def __notify(self):
        try:
            self.config.update_special_property(self.property_name)
        except:
            pass

    def __empty_prop(self, key):

        prop = BridgeSet(self.config)
        dict.__setitem__(self, key, prop)
        return prop

    def to_string(self, value: dict) -> str:
        out = []
        for jail_if in value:
            for bridge_if in self[jail_if]:
                out.append(f"{jail_if}:{bridge_if}")
        return " ".join(out)

    def __str__(self):
        return self.to_string(value=self)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import libzfs

import iocage.lib.Config.Dataset
import iocage.lib.Resource
import iocage.lib.errors


ZFS_PROPERTY_PREFIX = "org.freebsd.iocage:"


def is_iocage_property(name):

    return name.startswith(ZFS_PROPERTY_PREFIX)


def get_iocage_property_name(zfs_property_name):

    if is_iocage_property(zfs_property_name) is False:
        raise iocage.lib.errors.NotAnIocageZFSProperty(
            property_name=zfs_property_name
        )

    return zfs_property_name[len(ZFS_PROPERTY_PREFIX):]


class BaseConfigZFS(iocage.lib.Config.Dataset.DatasetConfig):

    config_type = "zfs"

    def read(self) -> dict:
        try:
            return self.map_input(self._read_properties())
        except:
            return {}

    def write(self, data: dict):
        """
        Writes changes to the config file
        """
        output_data = self.map_output(data)

        # ToDo: Delete unnecessary ZFS options
        # existing_property_names = list(
        #   map(lambda x: get_iocage_property_name(x),
        #     filter(
        #       lambda name: is_iocage_property(name),
        #       self.jail.dataset.properties
        #     )
        #   )
        # )
        # data_keys = list(self.data)
        # for existing_property_name in existing_property_names:
        #   if not existing_property_name in data_keys:
        #     pass

        for key, value in output_data.items():
            prop_name = f"{ZFS_PROPERTY_PREFIX}{key}"
            prop = libzfs.ZFSUserProperty(
                self._to_string(value)
            )
            self.dataset.properties[prop_name] = prop

    def map_input(self, data: dict) -> dict:
        parse_user_input = iocage.lib.helpers.parse_user_input
        return dict([(x, parse_user_input(y)) for (x, y) in data.items()])

    def map_output(self, data: dict) -> dict:
        output = {}
        for key, value in data.items():
            output[key] = self._to_string(value)
        return output

    def _to_string(self, value) -> str:
        return iocage.lib.helpers.to_string(
            value,
            true="on",
            false="off",
            none="none"
        )

    @property
    def exists(self) -> bool:
        """
        Returns True, when this configuration was found in the dataset
        """
        for prop in self.dataset.properties:
            if is_iocage_property(prop):
                return True

        return False

    def _read_properties(self) -> dict:
        data = {}
        for prop in self.dataset.properties:
            if is_iocage_property(prop):
                name = get_iocage_property_name(prop)
                data[name] = self.dataset.properties[prop].value
        return data


class ConfigZFS(BaseConfigZFS):

    def __init__(
        self,
        dataset: libzfs.ZFSDataset,
        **kwargs
    ) -> None:

        self._dataset = dataset
        BaseConfigZFS.__init__(self, **kwargs)

    @property
    def dataset(self) -> libzfs.ZFSDataset:
        return self._dataset


class ResourceConfigZFS(BaseConfigZFS):

    def __init__(
        self,
        resource: 'iocage.lib.Resource.Resource',
        **kwargs
    ) -> None:

        self.resource = resource
        BaseConfigZFS.__init__(self, **kwargs)

    @property
    def dataset(self) -> libzfs.ZFSDataset:
        return self.resource.dataset



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""get module for the cli."""
import typing
import click

import iocage.lib.errors
import iocage.lib.Host
import iocage.lib.Jail
import iocage.lib.Logger


@click.command(
    context_settings=dict(max_content_width=400,),
    name="get",
    help="Gets the specified property."
)
@click.pass_context
@click.argument("prop", required=True, default="")
@click.argument("jail", required=True, default="")
@click.option(
    "--all", "-a", "_all",
    help="Get all properties for the specified jail.",
    is_flag=True
)
@click.option(
    "--pool", "-p", "_pool",
    help="Get the currently activated zpool.",
    is_flag=True
)
@click.option(
    "--log-level", "-d",
    default="info"
)
def cli(ctx, prop, _all, _pool, jail, log_level):
    """Get a list of jails and print the property."""

    logger = ctx.parent.logger
    logger.print_level = log_level
    host = iocage.lib.Host.Host(logger=logger)

    if _pool is True:
        try:
            print(host.datasets.active_pool.name)
        except Exception:
            print("No active pool found")
        exit(1)

    if jail == "":
        prop = ""

    if jail == "defaults":
        source_resource = host.defaults
        lookup_method = _lookup_config_value
    else:
        lookup_method = _lookup_jail_value
        try:
            source_resource = iocage.lib.Jail.Jail(
                jail,
                host=host,
                logger=logger
            )
        except iocage.lib.errors.JailNotFound as e:
            exit(1)

    if _all is True:
        prop = None

    if prop == "all":
        prop = None

    if (prop is None) and (jail == "") and not _all:
        logger.error("Missing arguments property and jail")
        exit(1)
    elif (prop is not None) and (jail == ""):
        logger.error("Missing argument property name or -a/--all argument")
        exit(1)

    if prop:
        value = lookup_method(source_resource, prop)

        if value:
            print(value)
            return
        else:
            logger.error(f"Unknown property '{prop}'")
            exit(1)

    for key in source_resource.config.all_properties:
        if (prop is None) or (key == prop):
            value = source_resource.config.get_string(key)
            print_property(key, value)


def print_property(key, value):
    print(f"{key}:{value}")


def _lookup_config_value(resource, key: str) -> typing.Any:
    return iocage.lib.helpers.to_string(resource.config[key])


def _lookup_jail_value(resource, key: str) -> typing.Any:

    if key == "running":
        value = resource.running
    else:
        value = resource.getstring(key)

    return iocage.lib.helpers.to_string(value)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import locale
import os
import re
import signal
import subprocess as su
import sys

import click

from ..lib.Logger import Logger

logger = Logger()

click.core._verify_python3_env = lambda: None  # type: ignore
user_locale = os.environ.get("LANG", "en_US.UTF-8")
locale.setlocale(locale.LC_ALL, user_locale)

IOCAGE_CMD_FOLDER = os.path.abspath(os.path.dirname(__file__))

# @formatter:off
# Sometimes SIGINT won't be installed.
# http://stackoverflow.com/questions/40775054/capturing-sigint-using-keyboardinterrupt-exception-works-in-terminal-not-in-scr/40785230#40785230
signal.signal(signal.SIGINT, signal.default_int_handler)
# If a utility decides to cut off the pipe, we don't care (IE: head)
signal.signal(signal.SIGPIPE, signal.SIG_DFL)
# @formatter:on

try:
    su.check_call(["sysctl", "vfs.zfs.version.spa"],
                  stdout=su.PIPE, stderr=su.PIPE)
except su.CalledProcessError:
    logger.error(
        "ZFS is required to use iocage.\n"
        "Try calling 'kldload zfs' as root."
    )
    exit(1)


def print_events(generator):
    lines = {}
    for event in generator:

        if event.identifier is None:
            identifier = "generic"
        else:
            identifier = event.identifier

        if event.type not in lines:
            lines[event.type] = {}

        # output fragments
        running_indicator = "+" if (event.done or event.skipped) else "-"
        name = event.type
        if event.identifier is not None:
            name += f"@{event.identifier}"

        output = f"[{running_indicator}] {name}: "

        if event.message is not None:
            output += event.message
        else:
            output += event.get_state_string(
                done="OK",
                error="FAILED",
                skipped="SKIPPED",
                pending="..."
            )

        if event.duration is not None:
            output += " [" + str(round(event.duration, 3)) + "s]"

        # new line or update of previous
        if identifier not in lines[event.type]:
            # Indent if previous task is not finished
            lines[event.type][identifier] = logger.screen(
                output,
                indent=event.parent_count
            )
        else:
            lines[event.type][identifier].edit(
                output,
                indent=event.parent_count
            )


class IOCageCLI(click.MultiCommand):
    """
    Iterates in the 'cli' directory and will load any module's cli definition.
    """

    def list_commands(self, ctx):
        rv = []

        for filename in os.listdir(IOCAGE_CMD_FOLDER):
            if filename.endswith('.py') and \
                    not filename.startswith('__init__'):
                rv.append(re.sub(".py$", "", filename))
        rv.sort()

        return rv

    def get_command(self, ctx, name):
        ctx.print_events = print_events
        try:
            mod = __import__(f"iocage.cli.{name}", None, None, ["cli"])

            try:
                if mod.__rootcmd__ and "--help" not in sys.argv[1:]:
                    if len(sys.argv) != 1:
                        if os.geteuid() != 0:
                            app_name = mod.__name__.rsplit(".")[-1]
                            logger.error(
                                "You need to have root privileges"
                                f" to run {app_name}"
                            )
                            exit(1)
            except AttributeError:
                # It's not a root required command.
                pass
            return mod.cli
        except (ImportError, AttributeError):
            return


@click.option("--log-level", "-d", default=None)
@click.command(cls=IOCageCLI)
@click.version_option(version="0.2.12 09/17/2017", prog_name="ioc")
@click.pass_context
def cli(ctx, log_level):
    """A jail manager."""
    logger.print_level = log_level
    ctx.logger = logger



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import iocage.lib.errors
import iocage.lib.helpers


class Prompts:
    def __init__(self, host=None, logger=None):
        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.host = iocage.lib.helpers.init_host(self, host)

    def release(self):
        default = None
        available_releases = self.host.distribution.releases
        for i, available_release in enumerate(available_releases):
            if available_release.name == self.host.release_version:
                default = i
                print(
                    f"[{i}] \033[1m{available_release.annotated_name}\033[0m"
                )
            else:
                print(f"[{i}] {available_release.annotated_name}")

        if default is not None:
            default_release = available_releases[default]
            selection = input(
                # f"Release ({default_release.name}) [{default}]: "
                "\nType the number of the desired RELEASE\n"
                "Press [Enter] to fetch the default selection"
                f" ({default_release.name}) [{default}]: "
            )
        else:
            default_release = None
            selection = input("Your selection: ")

        if selection == "":
            if default_release is None:
                raise iocage.lib.errors.DefaultReleaseNotFound(
                    host_release_name=self.host.release_version,
                    logger=self.logger
                )
            return default_release
        else:
            return available_releases[int(selection)]



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""exec module for the cli."""
import click

import iocage.lib.Jail
import iocage.lib.Logger

__rootcmd__ = True


@click.command(context_settings=dict(ignore_unknown_options=True),
               name="exec", help="Run a command inside a specified jail.")
@click.pass_context
@click.option("--host_user", "-u", default="root",
              help="The host user to use.")
@click.option("--jail_user", "-U", help="The jail user to use.")
@click.argument("jail", required=True, nargs=1)
@click.argument("command", nargs=-1, type=click.UNPROCESSED)
@click.option("--log-level", "-d", default=None)
def cli(ctx, command, jail, host_user, jail_user, log_level):
    """Runs the command given inside the specified jail"""
    logger = ctx.parent.logger
    logger.print_level = log_level

    if jail.startswith("-"):
        logger.error("Please specify a jail first!")
        exit(1)

    user_command = " ".join(list(command))

    if jail_user:
        command = ["/bin/su", "-m",
                   jail_user, "-c", user_command]
    else:
        command = ["/bin/sh", "-c", user_command]

    jail = iocage.lib.Jail.Jail(jail, logger=logger)
    jail.update_jail_state()

    if not jail.exists:
        logger.error(f"The jail {jail.humanreadable_name} does not exist")
        exit(1)
    if not jail.running:
        logger.error(f"The jail {jail.humanreadable_name} is not running")
        exit(1)
    else:
        jail.passthru(command)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing

import iocage.lib.helpers
import iocage.lib.Config.Jail.BaseConfig

# MyPy
import iocage.lib.Jail

_usp = iocage.lib.Config.Jail.BaseConfig.BaseConfig.update_special_property


class JailConfig(iocage.lib.Config.Jail.BaseConfig.BaseConfig):

    legacy: bool = None
    jail: 'iocage.lib.Jail.JailGenerator' = None
    data: dict = {}

    def __init__(
        self,
        data: dict=None,
        jail: 'iocage.lib.Jail.JailGenerator'=None,
        new: bool=False,
        logger: 'iocage.lib.Logger.Logger'=None,
        host: 'iocage.lib.Host.HostGenerator'=None
    ) -> None:

        iocage.lib.Config.Jail.BaseConfig.BaseConfig.__init__(
            self,
            logger=logger
        )

        self.host = iocage.lib.helpers.init_host(self, host)

        if len(data.keys()) == 0:
            self.data = {
                "id": None
            }

        self.jail = jail

        # the name is used in many other variables and needs to be set first
        for key in ["id", "name", "uuid"]:
            if key in data.keys():
                self["id"] = data[key]
                break

        self.clone(data)

    def update_special_property(self, name: str) -> None:

        _usp(
            self,
            name=name
        )

        if (name == "ip6_addr") and (self.jail is not None):
            rc_conf = self.jail.rc_conf
            rc_conf["rtsold_enable"] = "accept_rtadv" in str(self["ip6_addr"])

    def _get_host_hostname(self):
        try:
            return self.data["host_hostname"]
        except KeyError:
            return self.jail.humanreadable_name

    def __getitem__(self, key: str) -> typing.Any:
        try:
            return super().__getitem__(key)
        except KeyError:
            # fall back to default
            return self.host.default_config[key]

    @property
    def all_properties(self) -> list:
        jail_config_properties = set(super().all_properties)
        default_config_properties = set(
            self.host.default_config.all_properties)
        return sorted(list(jail_config_properties | default_config_properties))



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import re
import uuid

import iocage.lib.Config.Jail.JailConfigProperties
import iocage.lib.errors
import iocage.lib.helpers


class BaseConfig(dict):
    """
    Represents an iocage jail's configuration

    A jail configuration can be loaded from various formats that were used
    by different versions of iocage. Technically it is possible to store
    set properties in deprecated formats, but this might break when using
    newer features than the legacy version of iocage supports. It is
    recommended to use the reading capabilities to migrate to the JSON
    config format.

    Supported Configuration Formats:

        JSON: (current)
            Since the Python 3 implementation iocage stored jail configs in
            a file called `config.json` within the jail's root dataset

        ZFS:
            iocage-legacy written in Bash used to save jail configurations
            as ZFS properties on the jail's root dataset. Due to poor
            performance and easier readability this format later was replaced
            with a file based config storage. Even though it is a deprecated
            format, libiocage is compatible to read a jail config from ZFS.

        UCL:
            Yet another deprecated configuration format, that libiocage also
            supports reading from. A legacy version of iocage used this format.

    Special Properties:

        Special properties are

    """

    special_properties: (
        'iocage.lib.Config.Jail.'
        'JailConfigProperties.JailConfigProperties'
    )
    data: dict = {}

    def __init__(
        self,
        logger: 'iocage.lib.Logger.Logger'=None,
    ) -> None:

        dict.__init__(self)

        self.logger = iocage.lib.helpers.init_logger(self, logger)

        tmp = iocage.lib.Config.Jail.JailConfigProperties
        self.special_properties = tmp.JailConfigProperties(  # type: ignore
            config=self,
            logger=self.logger
        )

    def clone(
        self,
        data: typing.Dict[str, typing.Any],
        skip_on_error: bool=False
    ) -> None:
        """
        Apply data from a data dictionary to the JailConfig

        Existing jail configuration is not emptied using.

        Args:

            data (dict):
                Dictionary containing the configuration to apply

            skip_on_error (bool):
                Passed to __setitem__

        """
        if len(data.keys()) == 0:
            return

        current_id = self["id"]
        for key, value in data.items():

            if (key in ["id", "name", "uuid"]) and (current_id is not None):
                value = current_id

            self.__setitem__(key, value, skip_on_error=skip_on_error)

    def read(self, data: dict) -> None:

        # ignore name/id/uuid in config if the jail already has a name
        if self["id"] is not None:
            data_keys = data.keys()
            if "name" in data_keys:
                del data["name"]
            if "id" in data_keys:
                del data["id"]
            if "uuid" in data_keys:
                del data["uuid"]

        self.clone(data)

    def update_special_property(self, name: str) -> None:
        self.data[name] = str(self.special_properties[name])

    def attach_special_property(self, name, special_property):
        self.special_properties[name] = special_property

    def _set_legacy(self, value, **kwargs) -> None:
        try:
            self.legacy = iocage.lib.helpers.parse_bool(value)
        except:
            self.legacy = False

    def _get_id(self) -> str:
        return self.data["id"]

    def _set_id(self, name: str, **kwargs):

        try:
            # We do not want to set the same name twice.
            # This can occur when the Jail is initialized
            # with it's name and the same name is read from
            # the configuration
            if self["id"] == name:
                return
        except:
            pass

        if name is None:
            self.data["id"] = None
            return

        allowed_characters_pattern = "([^A-z0-9\\._\\-]|\\^)"
        invalid_characters = re.findall(allowed_characters_pattern, name)
        if len(invalid_characters) > 0:
            msg = (
                f"Invalid character in name: "
                " ".join(invalid_characters)
            )
            self.logger.error(msg)

        is_valid_name = iocage.lib.helpers.validate_name(name)
        if is_valid_name is True:
            self.data["id"] = name
        else:
            try:
                self.data["id"] = str(uuid.UUID(name))  # legacy support
            except:
                raise iocage.lib.errors.InvalidJailName(logger=self.logger)

    def _get_name(self):
        return self._get_id()

    def _get_uuid(self):
        return self._get_id()

    def _get_type(self):

        if self["basejail"]:
            return "basejail"
        elif self["clonejail"]:
            return "clonejail"
        else:
            return "jail"

    def _set_type(self, value, **kwargs):

        if value == "basejail":
            self["basejail"] = True
            self["clonejail"] = False
            self.data["type"] = "jail"

        elif value == "clonejail":
            self["basejail"] = False
            self["clonejail"] = True
            self.data["type"] = "jail"

        else:
            self.data["type"] = value

    def _get_priority(self) -> int:
        return int(self.data["priority"])

    def _set_priority(self, value: typing.Union[int, str], **kwargs):
        self.data["priority"] = str(value)

    # legacy support
    def _get_tag(self) -> str:

        if self._has_legacy_tag is True:
            return self.data["tag"]

        try:
            return self["tags"][0]
        except KeyError:
            return None

    def _set_tag(self, value: str, **kwargs) -> None:

        if (self._has_legacy_tag is True) or ("tags" not in self.data.keys()):
            # store as deprecated `tag` for downwards compatibility
            # setting `tags` overrides usage of the deprecated `tag` property
            self.data["tag"] = value
            return

        tags = self["tags"]
        if value in tags:
            # remove the tag if it was existing
            del tags[tags.index(value)]

        tags.insert(0, value)
        self.data["tags"] = iocage.lib.helpers.to_string(tags)

    @property
    def _has_legacy_tag(self) -> bool:
        return "tag" in self.data.keys()

    def _get_tags(self) -> typing.List[str]:
        return iocage.lib.helpers.parse_list(self.data["tags"])

    def _set_tags(
        self,
        value: typing.Union[
            str,
            bool,
            int,
            typing.List[typing.Union[str, bool, int]]
        ],
        **kwargs
    ) -> None:

        data = iocage.lib.helpers.to_string(value)
        self.data["tags"] = data

        if self._has_legacy_tag is True:
            del self.data["tag"]

    def _get_basejail(self) -> bool:
        return iocage.lib.helpers.parse_bool(self.data["basejail"])

    def _set_basejail(self, value, **kwargs):
        self.data["basejail"] = self.stringify(value)

    def _get_clonejail(self) -> bool:
        return iocage.lib.helpers.parse_bool(self.data["clonejail"])

    def _set_clonejail(self, value, **kwargs):
        self.data["clonejail"] = self.stringify(value)

    def _get_defaultrouter(self):
        value = self.data['defaultrouter']
        return value if (value != "none" and value is not None) else None

    def _set_defaultrouter(self, value, **kwargs):
        if value is None:
            value = 'none'
        self.data['defaultrouter'] = value

    def _get_defaultrouter6(self):
        value = self.data['defaultrouter6']
        return value if (value != "none" and value is not None) else None

    def _set_defaultrouter6(self, value, **kwargs):
        if value is None:
            value = 'none'
        self.data['defaultrouter6'] = value

    def _get_vnet(self):
        return iocage.lib.helpers.parse_user_input(self.data["vnet"])

    def _set_vnet(self, value, **kwargs):
        self.data["vnet"] = iocage.lib.helpers.to_string(
            value,
            true="on",
            false="off"
        )

    def _get_jail_zfs_dataset(self):
        try:
            return self.data["jail_zfs_dataset"].split()
        except KeyError:
            return []

    def _set_jail_zfs_dataset(self, value, **kwargs):
        value = [value] if isinstance(value, str) else value
        self.data["jail_zfs_dataset"] = " ".join(value)

    def _get_jail_zfs(self):
        return iocage.lib.helpers.parse_user_input(
            self.data["jail_zfs"]
        )

    def _set_jail_zfs(self, value, **kwargs):
        parsed_value = iocage.lib.helpers.parse_user_input(value)
        if parsed_value is None:
            del self.data["jail_zfs"]
            return
        self.data["jail_zfs"] = iocage.lib.helpers.to_string(
            parsed_value,
            true="on",
            false="off"
        )

    def _get_cloned_release(self):
        try:
            return self.data["cloned_release"]
        except:
            return self["release"]

    def _get_basejail_type(self):

        # first see if basejail_type was explicitly set
        try:
            return self.data["basejail_type"]
        except:
            pass

        # if it was not, the default for is 'nullfs' if the jail is a basejail
        try:
            if self["basejail"]:
                return "nullfs"
        except:
            pass

        # otherwise the jail does not have a basejail_type
        return None

    def _get_login_flags(self):
        try:
            return JailConfigList(self.data["login_flags"].split())
        except KeyError:
            return JailConfigList(["-f", "root"])

    def _set_login_flags(self, value, **kwargs):
        if value is None:
            try:
                del self.data["login_flags"]
            except:
                pass
        else:
            if isinstance(value, list):
                self.data["login_flags"] = " ".join(value)
            elif isinstance(value, str):
                self.data["login_flags"] = value
            else:
                raise iocage.lib.errors.InvalidJailConfigValue(
                    property_name="login_flags",
                    logger=self.logger
                )

    def _get_host_hostuuid(self):
        try:
            return self.data["host_hostuuid"]
        except KeyError:
            return self["id"]

    def get_string(self, key):
        return self.stringify(self.__getitem__(key))

    def _skip_on_error(self, **kwargs):
        """
        A helper to resolve skip_on_error attribute
        """
        try:
            return kwargs["skip_on_error"] is True
        except AttributeError:
            return False

    def __getitem_user(self, key):

        # passthrough existing properties
        try:
            return self.__getattribute__(key)
        except:
            pass

        is_special_property = self.special_properties.is_special_property(key)
        is_existing = key in self.data.keys()
        if (is_special_property and is_existing) is True:
            return self.special_properties.get_or_create(key)

        # data with mappings
        get_method = None
        try:
            get_method = self.__getattribute__(f"_get_{key}")
            return get_method()
        except AttributeError:
            if get_method is not None:
                raise
            pass

        # plain data attribute
        try:
            return self.data[key]
        except:
            pass

        raise KeyError(f"User defined property not found: {key}")

    def __getitem__(self, key: str) -> typing.Any:

        try:
            return self.__getitem_user(key)
        except KeyError:
            pass

        raise KeyError(f"Item not found: {key}")

    def __delitem__(self, key):
        del self.data[key]

    def __setitem__(
        self,
        key: str,
        value: typing.Any,
        **kwargs
    ):

        parsed_value = iocage.lib.helpers.parse_user_input(value)

        if self.special_properties.is_special_property(key):
            special_property = self.special_properties.get_or_create(key)
            special_property.set(value)
            self.update_special_property(key)
            return

        try:
            setter_method = self.__getattribute__(f"_set_{key}")
            if setter_method is None:
                return None
            else:
                return setter_method(parsed_value, **kwargs)
        except AttributeError:
            pass

        self.data[key] = parsed_value

    def set(self, key: str, value, **kwargs) -> bool:
        """
        Set a JailConfig property

        Args:

            key:
                The jail config property name

            value:
                Value to set the property to

            **kwargs:
                Arguments from **kwargs are passed to setter functions

        Returns:

            bool: True if the JailConfig was changed
        """

        try:
            hash_before = str(self.__getitem_user(key)).__hash__()
        except Exception:
            hash_before = None
            pass

        self.__setitem__(key, value, **kwargs)

        try:
            hash_after = str(self.__getitem_user(key)).__hash__()
        except Exception:
            hash_after = None
            pass

        return (hash_before != hash_after)

    @property
    def user_data(self) -> dict:
        return self.data

    def __str__(self) -> str:
        return iocage.lib.helpers.to_json(self.user_data)

    def __dir__(self) -> list:

        properties = set()
        props = dict.__dir__(self)  # type: ignore
        for prop in props:
            if not prop.startswith("_"):
                properties.add(prop)

        for key in self.data.keys():
            properties.add(key)

        return list(properties)

    @property
    def all_properties(self) -> list:
        return sorted(self.data.keys())

    def stringify(self, value):
        parsed_input = iocage.lib.helpers.parse_user_input(value)
        return iocage.lib.helpers.to_string(parsed_input)


class JailConfigList(list):

    def __str__(self) -> str:
        return " ".join(self)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import subprocess
from hashlib import md5

import iocage.lib.NetworkInterface
import iocage.lib.errors
import iocage.lib.helpers


class Network:
    def __init__(self, jail,
                 nic="vnet0",
                 ipv4_addresses=[],
                 ipv6_addresses=[],
                 mtu=1500,
                 bridges=None,
                 logger=None):

        self.logger = iocage.lib.helpers.init_logger(self, logger)

        if bridges is not None:
            if not isinstance(bridges, list):
                raise iocage.lib.errors.InvalidNetworkBridge(
                    reason="Expected None or list of strings",
                    logger=self.logger
                )

        self.vnet = True
        self.bridges = bridges
        self.jail = jail
        self.nic = nic
        self.mtu = mtu
        self.ipv4_addresses = ipv4_addresses
        self.ipv6_addresses = ipv6_addresses

    def setup(self):
        if self.vnet:
            if not self.bridges or len(self.bridges) == 0:
                raise iocage.lib.errors.VnetBridgeMissing(
                    logger=self.logger
                )

            jail_if, host_if = self.__create_vnet_iface()

    def teardown(self):
        if self.vnet:
            # down host_if
            iocage.lib.NetworkInterface.NetworkInterface(
                name=self.nic_local_name,
                extra_settings=["down"],
                logger=self.logger
            )

    @property
    def nic_local_name(self):
        self.jail.require_jail_running(silent=True)
        return f"{self.nic}:{self.jail.jid}"

    @property
    def nic_local_description(self):
        return f"associated with jail: {self.jail.humanreadable_name}"

    def __create_vnet_iface(self):

        # create new epair interface
        epair_a_cmd = ["ifconfig", "epair", "create"]
        epair_a = subprocess.Popen(
            epair_a_cmd, stdout=subprocess.PIPE, shell=False).communicate()[0]
        epair_a = epair_a.decode("utf-8").strip()
        epair_b = f"{epair_a[:-1]}b"

        mac_a, mac_b = self.__generate_mac_address_pair()

        host_if = iocage.lib.NetworkInterface.NetworkInterface(
            name=epair_a,
            mac=mac_a,
            mtu=self.mtu,
            description=self.nic_local_description,
            rename=self.nic_local_name,
            logger=self.logger
        )

        # add host_if to bridges
        for bridge in self.bridges:
            iocage.lib.NetworkInterface.NetworkInterface(
                name=bridge,
                addm=self.nic_local_name,
                extra_settings=["up"],
                logger=self.logger
            )

        # up host_if
        iocage.lib.NetworkInterface.NetworkInterface(
            name=self.nic_local_name,
            extra_settings=["up"],
            logger=self.logger
        )

        # assign epair_b to jail
        self.__assign_vnet_iface_to_jail(epair_b, self.jail.identifier)

        jail_if = iocage.lib.NetworkInterface.NetworkInterface(
            name=epair_b,
            mac=mac_b,
            mtu=self.mtu,
            rename=self.nic,
            jail=self.jail,
            extra_settings=["up"],
            ipv4_addresses=self.ipv4_addresses,
            ipv6_addresses=self.ipv6_addresses,
            logger=self.logger
        )

        return jail_if, host_if

    def __assign_vnet_iface_to_jail(self, nic, jail_name):
        iocage.lib.NetworkInterface.NetworkInterface(
            name=nic,
            vnet=jail_name,
            logger=self.logger
        )

    def __generate_mac_bytes(self):
        m = md5()
        m.update(self.jail.name.encode("utf-8"))
        m.update(self.nic.encode("utf-8"))
        prefix = self.jail.config["mac_prefix"]
        return f"{prefix}{m.hexdigest()[0:12-len(prefix)]}"

    def __generate_mac_address_pair(self):
        mac_a = self.__generate_mac_bytes()
        mac_b = hex(int(mac_a, 16) + 1)[2:].zfill(12)
        return mac_a, mac_b



import typing
import iocage.lib.Config.Jail.Properties.Addresses
import iocage.lib.Config.Jail.Properties.Interfaces
import iocage.lib.Config.Jail.Properties.Resolver

_Config = iocage.lib.Config

CLASSES: dict = dict(
    ip4_addr=iocage.lib.Config.Jail.Properties.Addresses.AddressesProp,
    ip6_addr=iocage.lib.Config.Jail.Properties.Addresses.AddressesProp,
    interfaces=iocage.lib.Config.Jail.Properties.Interfaces.InterfaceProp,
    resolver=iocage.lib.Config.Jail.Properties.Resolver.ResolverProp
)


def init_property(property_name: str, **kwargs) -> typing.Union[
    iocage.lib.Config.Jail.Properties.Addresses.AddressesProp,
    iocage.lib.Config.Jail.Properties.Interfaces.InterfaceProp,
    iocage.lib.Config.Jail.Properties.Resolver.ResolverProp
]:

    target_class = CLASSES[property_name]
    out = target_class(
        property_name=property_name,
        **kwargs
    )
    return out



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import iocage.lib.errors


class ZFSShareStorage:
    @property
    def zfs_datasets(self):
        return self._get_zfs_datasets(auto_create=self.auto_create)

    def mount_zfs_shares(self, auto_create=False):
        self.logger.log("Mounting ZFS shares")
        self._mount_procfs()
        self._mount_jail_datasets(auto_create=auto_create)

    def _get_zfs_datasets(self, auto_create=None):
        dataset_names = self.jail.config["jail_zfs_dataset"]

        auto_create = self.auto_create if auto_create is None else auto_create

        datasets = set()
        for name in dataset_names:

            zpool = None
            try:
                zpool = self._get_pool_from_dataset_name(name)
            except:
                pass

            pool_name = f"{self.jail.pool_name}/{name}"
            try:
                # legacy support (datasets not prefixed with pool/)
                zpool = self._get_pool_from_dataset_name(pool_name)
                name = f"{self.jail.pool_name}/{name}"
            except:
                pass

            try:
                if auto_create:
                    zpool.create(name, {}, create_ancestors=True)
            except:
                pass

            try:
                dataset = self.zfs.get_dataset(name)
                datasets.add(dataset)
            except:
                raise iocage.lib.errors.DatasetNotAvailable(
                    dataset_name=name,
                    logger=self.logger
                )

        return datasets

    def _mount_jail_dataset(self, dataset_name):
        self.jail.exec(['zfs', 'mount', dataset_name])

    def _mount_jail_datasets(self, auto_create=None):

        auto_create = self.auto_create if auto_create is None else (
            auto_create is True)

        if self.safe_mode:
            self._require_datasets_exist_and_jailed()

        for dataset in self.zfs_datasets:

            self._unmount_local(dataset)

            # ToDo: bake jail feature into py-libzfs
            iocage.lib.helpers.exec(
                ["zfs", "jail", self.jail.identifier, dataset.name])

            if dataset.properties['mountpoint']:
                for child in list(dataset.children):
                    self._ensure_dataset_exists(child)
                    self._mount_jail_dataset(child.name)

    def _get_pool_name_from_dataset_name(self, dataset_name):
        return dataset_name.split("/", maxsplit=1)[0]

    def _get_pool_from_dataset_name(self, dataset_name):
        target_pool_name = self._get_pool_name_from_dataset_name(dataset_name)
        for zpool in list(self.zfs.pools):
            if zpool.name == target_pool_name:
                return zpool

        # silent exception, no logger defined
        raise iocage.lib.errors.ZFSPoolUnavailable(
            pool_name=target_pool_name
        )

    def _require_datasets_exist_and_jailed(self):
        existing_datasets = self.get_zfs_datasets(auto_create=False)
        for existing_dataset in existing_datasets:
            if existing_dataset.properties["jailed"] != "on":
                raise iocage.lib.errors.DatasetNotJailed(
                    dataset=existing_dataset,
                    logger=self.logger
                )



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import json

import iocage.lib.Config
import iocage.lib.Config.File
import iocage.lib.Config.Resource.ResourceConfig
import iocage.lib.helpers


class ConfigJSON(iocage.lib.Config.File.ConfigFile):

    config_type = "json"

    def map_input(self, data: typing.TextIO) -> dict:
        return json.load(data)

    def map_output(self, data: dict) -> str:
        return iocage.lib.helpers.to_json(data)


class ResourceConfigJSON(
    iocage.lib.Config.Resource.ResourceConfig.ResourceConfig,
    ConfigJSON
):
    pass



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import iocage.lib.StandaloneJailStorage
import iocage.lib.helpers


class NullFSBasejailStorage:
    """
    Iocage NullFS storage backend

    This backend is used by NullFS basejails.
    """

    def apply(self, release=None):
        NullFSBasejailStorage._create_nullfs_directories(self)

    def setup(self, release):
        iocage.lib.StandaloneJailStorage.StandaloneJailStorage.setup(
            self, release)

    def umount_nullfs(self):
        """
        Unmount all NullFS mounts from fstab

        In preparation of starting the jail with NullFS mounts all mountpoints
        that are listed in fstab need to be unmounted
        """
        with open(f"{self.jail.path}/fstab") as f:
            mounts = []
            for mount in f.read().splitlines():
                try:
                    mounts.append(mount.replace("\t", " ").split(" ")[1])
                except:
                    pass

            if (len(mounts) > 0):
                try:
                    iocage.lib.helpers.exec(["umount"] + mounts)
                except:
                    # in case directories were not mounted
                    pass

    def _create_nullfs_directories(self):
        basedirs = iocage.lib.helpers.get_basedir_list(
            distribution_name=self.jail.host.distribution.name
        ) + ["dev", "etc"]

        for basedir in basedirs:
            self.create_jail_mountpoint(basedir)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANYw
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""list module for the cli."""
import click
import json
import texttable
import typing

import iocage.lib.Logger
import iocage.lib.Host
import iocage.lib.Resource
import iocage.lib.Jails
import iocage.lib.Releases


supported_output_formats = ['table', 'csv', 'list', 'json']


@click.command(name="list", help="List a specified dataset type, by default"
                                 " lists all jails.")
@click.pass_context
@click.option("--release", "--base", "-r", "-b", "dataset_type",
              flag_value="base", help="List all bases.")
@click.option("--template", "-t", "dataset_type", flag_value="template",
              help="List all templates.")
@click.option("--long", "-l", "_long", is_flag=True, default=False,
              help="Show the full uuid and ip4 address.")
@click.option("--remote", "-R", is_flag=True, help="Show remote's available "
                                                   "RELEASEs.")
@click.option("--plugins", "-P", is_flag=True, help="Show available plugins.")
@click.option("--sort", "-s", "_sort", default=None, nargs=1,
              help="Sorts the list by the given type")
@click.option("--quick", "-q", is_flag=True, default=False,
              help="Lists all jails with less processing and fields.")
@click.option("--output", "-o", default=None)
@click.option("--output-format", "-f", default="table",
              type=click.Choice(supported_output_formats))
@click.option("--header/--no-header", "-H/-NH", is_flag=True, default=True,
              help="Show or hide column name heading.")
@click.argument("filters", nargs=-1)
def cli(ctx, dataset_type, header, _long, remote, plugins,
        _sort, quick, output, output_format, filters):
    logger = ctx.parent.logger

    host = iocage.lib.Host.Host(logger=logger)

    if remote and not plugins:

        available_releases = host.distribution.releases
        for available_release in available_releases:
            logger.screen(available_release.name)
        return

    if plugins and remote:
        raise iocage.lib.errors.MissingFeature("Plugins", plural=True)

    if output is not None and _long is True:
        logger.error("--output and --long can't be used together")
        exit(1)

    if output_format != "table" and _sort is not None:
        # Sorting destroys the ability to stream generators
        # ToDo: Figure out if we need to sort other output formats as well
        raise Exception("Sorting only allowed for tables")

    # empty filters will match all jails
    if len(filters) == 0:
        filters += ("*",)

    if dataset_type == "base":
        resources_class = iocage.lib.Releases.ReleasesGenerator
        columns = ["name"]
    else:
        resources_class = iocage.lib.Jails.JailsGenerator
        columns = _list_output_comumns(output, _long)

    resources = resources_class(
        logger=logger,
        host=host,
        filters=filters  # ToDo: allow quoted whitespaces from user input
    )

    if output_format == "list":
        _print_list(resources, columns, header, "\t")
    elif output_format == "csv":
        _print_list(resources, columns, header, ";")
    elif output_format == "json":
        _print_json(resources, columns)
    else:
        _print_table(resources, columns, header, _sort)


def _print_table(
    resources: typing.Generator[
        iocage.lib.Resource.ListableResource,
        None,
        None
    ],
    columns: list,
    show_header: bool,
    sort_key: str=None
) -> None:

    table = texttable.Texttable(max_width=0)
    table.set_cols_dtype(["t"] * len(columns))

    table_head = (list(x.upper() for x in columns))
    table_data = []

    try:
        sort_index = columns.index(sort_key)
    except ValueError:
        sort_index = None

    for resource in resources:
        table_data.append(_lookup_resource_values(resource, columns))

    if sort_index is not None:
        table_data.sort(key=lambda x: x[sort_index])

    if show_header:
        table.add_rows([table_head] + table_data)
    else:
        table.add_rows(table_data)

    print(table.draw())


def _print_list(
    resources: typing.Generator[
        iocage.lib.Jails.JailsGenerator,
        None,
        None
    ],
    columns: list,
    show_header: bool,
    separator: str=";"
) -> None:

    if show_header is True:
        print(separator.join(columns).upper())

    for resource in resources:
        print(separator.join(_lookup_resource_values(resource, columns)))


def _print_json(
    resources: typing.Generator[
        iocage.lib.Jails.JailsGenerator,
        None,
        None
    ],
    columns: list,
    **json_dumps_args
):

    if "indent" not in json_dumps_args.keys():
        json_dumps_args["indent"] = 2

    if "sort_keys" not in json_dumps_args.keys():
        json_dumps_args["sort_keys"] = True

    output = []

    for resource in resources:
        output.append(dict(zip(
            columns,
            _lookup_resource_values(resource, columns)
        )))

    print(json.dumps(output, **json_dumps_args))


def _lookup_resource_values(resource, columns) -> typing.List[str]:
    return list(map(
        lambda column: resource.getstring(column),
        columns
    ))


def _list_output_comumns(
    user_input: str="",
    long_mode: bool=False
) -> list:

    if user_input:
        return user_input.strip().split(',')
    else:
        columns = ["jid", "name"]

        if long_mode is True:
            columns += [
                "running",
                "release",
                "ip4.addr",
                "ip6.addr"
            ]
        else:
            columns += [
                "running",
                "release",
                "ip4.addr"
            ]

        return columns



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""console module for the cli."""

import click

import iocage.lib.Jail
import iocage.lib.Logger

__rootcmd__ = True


@click.command(name="console", help="Login to a jail.")
@click.pass_context
@click.argument("jail")
@click.option("--log-level", "-d", default=None)
def cli(ctx, jail, log_level):
    """
    Runs jexec to login into the specified jail.
    """
    logger = ctx.parent.logger
    logger.print_level = log_level

    jail = iocage.lib.Jail.Jail(jail, logger=logger)
    jail.update_jail_state()

    if not jail.exists:
        logger.error(f"The jail {jail.humanreadable_name} does not exist")
        exit(1)
    if not jail.running:
        logger.error(f"The jail {jail.humanreadable_name} is not running")
        exit(1)
    else:
        jail.exec_console()



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import libzfs

import iocage.lib.Logger
import iocage.lib.helpers
import iocage.lib.errors


def get_zfs(
    logger: 'iocage.lib.Logger.Logger'=None,
    history: bool=True,
    history_prefix: str="<iocage>"
):
    zfs = ZFS(history=history, history_prefix=history_prefix)
    zfs.logger = iocage.lib.helpers.init_logger(zfs, logger)
    return zfs


class ZFS(libzfs.ZFS):

    logger: iocage.lib.Logger.Logger = None

    def create_dataset(
        self,
        dataset_name: str,
        **kwargs
    ) -> libzfs.ZFSDataset:

        pool = self.get_pool(dataset_name)
        pool.create(dataset_name, kwargs, create_ancestors=True)

        dataset = self.get_dataset(dataset_name)
        dataset.mount()
        return dataset

    def get_or_create_dataset(
        self,
        dataset_name: str,
        **kwargs
    ) -> libzfs.ZFSDataset:

        try:
            return self.get_dataset(dataset_name)
        except:
            pass

        return self.create_dataset(dataset_name, **kwargs)

    def get_pool(self, name: str) -> libzfs.ZFSPool:
        pool_name = name.split("/")[0]
        for pool in self.pools:
            if pool.name == pool_name:
                return pool
        raise iocage.lib.errors.ZFSPoolUnavailable(
            pool_name=pool_name,
            logger=self.logger
        )



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""The main CLI for ioc."""
import libzfs

import iocage.lib.Config.Jail.File.RCConf
import iocage.lib.Resource

# MyPy
import iocage.lib.Config.Jail.JailConfig


class LaunchableResource(iocage.lib.Resource.Resource):

    _rc_conf: 'iocage.lib.Config.Jail.File.RCConf.RCConf' = None
    config: 'iocage.lib.Config.Jail.JailConfig.JailConfig' = None

    def create_resource(self) -> None:
        """
        Creates the root dataset
        """
        iocage.lib.Resource.Resource.create_resource(self)
        self.zfs.create_dataset(self.root_dataset_name)

    @property
    def root_path(self):
        return self.root_dataset.mountpoint

    @property
    def root_dataset(self) -> libzfs.ZFSDataset:
        # ToDo: Memoize root_dataset
        return self.get_dataset("root")

    @property
    def root_dataset_name(self) -> str:
        return f"{self.dataset_name}/root"

    @property
    def dataset_name(self) -> str:
        raise NotImplementedError(
            "This needs to be implemented by the inheriting class"
        )

    @dataset_name.setter
    def dataset_name(self, value: str) -> None:
        raise NotImplementedError(
            "This needs to be implemented by the inheriting class"
        )

    @property
    def dataset(self) -> libzfs.ZFSDataset:
        if self._dataset is None:
            self._dataset = self.zfs.get_dataset(self.dataset_name)

        return self._dataset

    @dataset.setter
    def dataset(self, value: libzfs.ZFSDataset):
        self._set_dataset(value)

    @property
    def rc_conf(self) -> 'iocage.lib.Config.Jail.File.RCConf.RCConf':
        if self._rc_conf is None:
            self._rc_conf = iocage.lib.Config.Jail.File.RCConf.RCConf(
                resource=self,
                logger=self.logger
            )
        return self._rc_conf



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import libzfs

import iocage.lib.errors
import iocage.lib.helpers


class Datasets:
    ZFS_POOL_ACTIVE_PROPERTY = "org.freebsd.ioc:active"

    def __init__(self, root=None, pool=None, zfs=None, logger=None):
        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.zfs = iocage.lib.helpers.init_zfs(self, zfs)
        self._datasets = {}

        if isinstance(root, libzfs.ZFSDataset):
            self.root = root
            return

        if isinstance(pool, libzfs.ZFSPool):
            self.root = self._get_or_create_dataset(
                "iocage",
                root_name=pool.name,
                pool=pool
            )
            return

        active_pool = self.active_pool

        if active_pool is None:
            raise iocage.lib.errors.IocageNotActivated(logger=self.logger)
        else:
            self.root = self.zfs.get_dataset(f"{active_pool.name}/iocage")

    @property
    def active_pool(self):
        for pool in self.zfs.pools:
            if self._is_pool_active(pool):
                return pool
        return None

    @property
    def releases(self):
        return self._get_or_create_dataset("releases")

    @property
    def base(self):
        return self._get_or_create_dataset("base")

    @property
    def jails(self):
        return self._get_or_create_dataset("jails")

    @property
    def logs(self):
        return self._get_or_create_dataset("log")

    def activate(self, mountpoint=None):
        self.activate_pool(self.root.pool, mountpoint)

    def activate_pool(self, pool, mountpoint=None):

        if self._is_pool_active(pool):
            msg = f"ZFS pool '{pool.name}' is already active"
            self.logger.warn(msg)

        if not isinstance(pool, libzfs.ZFSPool):
            raise iocage.lib.errors.ZFSPoolInvalid("cannot activate")

        if pool.status == "UNAVAIL":
            raise iocage.lib.errors.ZFSPoolUnavailable(pool.name)

        other_pools = filter(lambda x: x.name != pool.name, self.zfs.pools)
        for other_pool in other_pools:
            self._deactivate_pool(other_pool)

        self._activate_pool(pool)

        root_dataset_args = {
            "pool": pool
        }

        if mountpoint is not None:
            root_dataset_args["mountpoint"] = mountpoint

        self.root = self._get_or_create_dataset(
            "iocage",
            **root_dataset_args
        )

    def _is_pool_active(self, pool):
        return iocage.lib.helpers.parse_user_input(self._get_pool_property(
            pool,
            self.ZFS_POOL_ACTIVE_PROPERTY
        ))

    def _get_pool_property(self, pool, prop):
        try:
            return pool.root_dataset.properties[prop].value
        except (KeyError, ValueError):
            return None

    def _get_dataset_property(self, dataset, prop):
        try:
            return dataset.properties[prop].value
        except:
            return None

    def _activate_pool(self, pool):
        self._set_pool_activation(pool, True)

    def _deactivate_pool(self, pool):
        self._set_pool_activation(pool, False)

    def _set_pool_activation(self, pool, state):
        value = "yes" if state is True else "no"
        self._set_zfs_property(
            pool.root_dataset,
            self.ZFS_POOL_ACTIVE_PROPERTY,
            value
        )

    def _set_zfs_property(self, dataset, name, value):
        current_value = self._get_dataset_property(dataset, name)
        if current_value != value:
            self.logger.verbose(
                f"Set ZFS property {name}='{value}'"
                f" on dataset '{dataset.name}'"
            )
            dataset.properties[name] = libzfs.ZFSUserProperty(value)

    def _get_or_create_dataset(self,
                               name,
                               root_name=None,
                               pool=None,
                               mountpoint=None):

        if not iocage.lib.helpers.validate_name(name):
            raise NameError(f"Invalid 'name' for Dataset: {name}")

        try:
            return self.datasets[name]
        except (AttributeError, KeyError):
            pass

        if root_name is None:
            root_name = self.root.name

        if pool is None:
            pool = self.root.pool

        name = f"{root_name}/{name}"
        try:
            dataset = self.zfs.get_dataset(name)
        except:
            pool.create(name, {})
            dataset = self.zfs.get_dataset(name)

            if mountpoint is not None:
                mountpoint_property = libzfs.ZFSUserProperty(mountpoint)
                dataset.properties["mountpoint"] = mountpoint_property

            dataset.mount()
        self._datasets[name] = dataset

        return dataset



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""activate module for the cli."""
import click

import iocage.lib.Datasets
import iocage.lib.Logger
import iocage.lib.ZFS

__rootcmd__ = True


@click.command(name="activate", help="Set a zpool active for iocage usage.")
@click.pass_context
@click.argument("zpool")
@click.option("--mountpoint", "-m", default="/iocage")
def cli(ctx, zpool, mountpoint):
    """
    Calls ZFS set to change the property org.freebsd.ioc:active to yes.
    """
    logger = ctx.parent.logger
    zfs = iocage.lib.ZFS.get_zfs()
    iocage_pool = None

    for pool in zfs.pools:
        if pool.name == zpool:
            iocage_pool = pool

    if iocage_pool is None:
        logger.error(f"ZFS pool '{zpool}' not found")
        exit(1)

    try:
        datasets = iocage.lib.Datasets.Datasets(
            pool=iocage_pool,
            zfs=zfs,
            logger=logger
        )
        datasets.activate(mountpoint=mountpoint)
        logger.log(f"ZFS pool '{zpool}' activated")
    except:
        exit(1)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""stop module for the cli."""
import click

import iocage.lib.Jail
import iocage.lib.Jails
import iocage.lib.Releases
import iocage.lib.Logger
import iocage.lib.Resource

__rootcmd__ = True


@click.command(name="destroy", help="Destroy specified resource")
@click.pass_context
@click.option("--force", "-f", default=False, is_flag=True,
              help="Destroy the jail without warnings or more user input.")
@click.option("--release", "-r", default=False, is_flag=True,
              help="Destroy a specified RELEASE dataset.")
@click.option("--recursive", "-R", default=False, is_flag=True,
              help="Bypass the children prompt, best used with --force (-f).")
@click.option("--download", "-d", default=False, is_flag=True,
              help="Destroy the download dataset of the specified RELEASE as"
                   " well.")
@click.argument("filters", nargs=-1)
def cli(ctx, force, release, recursive, download, filters):
    """
    Looks for the jail supplied and passes the uuid, path and configuration
    location to stop_jail.
    """
    logger = ctx.parent.logger
    host = iocage.lib.Host.Host(logger=logger)

    if len(filters) == 0:
        logger.error("No filter specified - cannot select a target to delete")
        exit(1)

    if release is True:
        resources_class = iocage.lib.Releases.ReleasesGenerator
    else:
        resources_class = iocage.lib.Jails.JailsGenerator

    resources = resources_class(
        filters=filters,
        host=host,
        logger=logger
    )

    if len(list(resources)) == 0:
        print(filters)
        logger.error("No target matched your input")
        exit(1)

    failed_items = []

    for item in resources:

        old_mountpoint = item.dataset.mountpoint

        # ToDo: generalize and move this Method from Jail to LaunchableResource
        if isinstance(item, iocage.lib.Jail.JailGenerator):
            try:
                item.require_jail_stopped()
            except:
                logger.log(
                    "Jail '{item.name}' needs to be stopped before destruction"
                )
                failed_items.append(item)
                continue

        item.destroy()
        logger.screen(f"{old_mountpoint} destroyed")

    if len(failed_items) > 0:
        exit(1)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import json
import re
import subprocess
import uuid

import iocage.lib.errors
import iocage.lib.Datasets
import iocage.lib.Host
import iocage.lib.Logger
import iocage.lib.ZFS

# MyPy
import iocage.lib.Types


def init_zfs(
        self: typing.Any,
        zfs: 'iocage.lib.ZFS.ZFS'=None
) -> 'iocage.lib.ZFS.ZFS':

    try:
        return self.zfs
    except:
        pass

    if (zfs is not None) and isinstance(zfs, iocage.lib.ZFS.ZFS):
        object.__setattr__(self, 'zfs', zfs)
    else:
        new_zfs = iocage.lib.ZFS.get_zfs(logger=self.logger)
        object.__setattr__(self, 'zfs', new_zfs)

    return self.zfs


def init_host(
        self: typing.Any,
        host: 'iocage.lib.Host.HostGenerator'=None
) -> 'iocage.lib.Host.HostGenerator':

    try:
        return self.host
    except:
        pass

    if host:
        return host

    return iocage.lib.Host.HostGenerator(
        logger=self.logger,
        zfs=self.zfs
    )


def init_logger(
        self: typing.Any,
        logger: 'iocage.lib.Logger.Logger'=None
) -> 'iocage.lib.Logger.Logger':

    try:
        return self.logger
    except:
        pass

    if logger is not None:
        object.__setattr__(self, 'logger', logger)
        return logger
    else:
        try:
            return self.logger
        except:
            new_logger = iocage.lib.Logger.Logger()
            object.__setattr__(self, 'logger', new_logger)
            return new_logger


def exec(
    command: typing.List[str],
    logger: 'iocage.lib.Logger.Logger'=None,
    ignore_error: bool=False,
    **subprocess_args
) -> typing.Tuple[subprocess.Popen, str, str]:

    if isinstance(command, str):
        command = [command]

    command_str = " ".join(command)

    if logger:
        logger.log(f"Executing: {command_str}", level="spam")

    subprocess_args["stdout"] = subprocess_args.get("stdout", subprocess.PIPE)
    subprocess_args["stderr"] = subprocess_args.get("stderr", subprocess.PIPE)

    child = subprocess.Popen(
        command,
        shell=False,
        **subprocess_args
    )

    stdout, stderr = child.communicate()
    stdout = stdout.decode("UTF-8").strip()
    stderr = stderr.decode("UTF-8").strip()

    if logger and stdout:
        logger.spam(_prettify_output(stdout))

    if child.returncode > 0:

        if logger:
            log_level = "spam" if ignore_error else "warn"
            logger.log(
                f"Command exited with {child.returncode}: {command_str}",
                level=log_level
            )
            if stderr:
                logger.log(_prettify_output(stderr), level=log_level)

        if ignore_error is False:
            raise iocage.lib.errors.CommandFailure(
                returncode=child.returncode,
                logger=logger
            )

    return child, stdout, stderr


def _prettify_output(output: str) -> str:
    return "\n".join(map(
        lambda line: f"    {line}",
        output.split("\n")
    ))


def to_humanreadable_name(name: str) -> str:
    try:
        uuid.UUID(name)
        return str(name)[:8]
    except (TypeError, ValueError):
        return name


# helper function to validate names
_validate_name = re.compile(r'[a-z0-9][a-z0-9\.\-_]{0,31}', re.I)


def validate_name(name: str) -> bool:
    return bool(_validate_name.fullmatch(name))


def parse_none(data: typing.Any) -> None:
    if data is None:
        return None

    if data in ["none", "-", ""]:
        return None

    raise TypeError("Value is not None")


def parse_list(data: typing.Union[str, typing.List[str]]) -> list:
    """
    Transforms a comma separated string into a list
    """
    # ToDo: escaped commas
    return data if isinstance(data, list) else data.split(",")


def parse_bool(data: typing.Optional[typing.Union[str, bool]]) -> bool:
    """
    try to parse booleans from strings

    On success, it returns the parsed boolean on failure it raises a TypeError.

    Usage:
        >>> parse_bool("YES")
        True
        >>> parse_bool("false")
        False
        >>> parse_bool("/etc/passwd")
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        TypeError: Not a boolean value
    """

    if isinstance(data, bool):
        return data
    if isinstance(data, str):
        val = data.lower()
        if val in ["yes", "true", "on"]:
            return True
        elif val in ["no", "false", "off"]:
            return False

    raise TypeError("Value is not a boolean")


def parse_user_input(data: typing.Optional[typing.Union[str, bool]]):
    """
    uses parse_bool() to partially return Boolean and NoneType values
    All other types as returned as-is

    >>> parse_user_input("YES")
    True
    >>> parse_user_input("false")
    False
    >>> parse_user_input("notfalse")
    'notfalse'
    >>> parse_user_input(8.4)
    8.4
    """

    try:
        return parse_bool(data)
    except TypeError:
        pass

    try:
        return parse_none(data)
    except TypeError:
        pass

    return data


def to_json(data: dict) -> str:
    output_data = {}
    for key, value in data.items():
        output_data[key] = to_string(
            value,
            true="yes",
            false="no",
            none="none"
        )
    return json.dumps(output_data, sort_keys=True, indent=4)


def to_string(
    data: typing.Union[
        str,
        bool,
        int,
        None,
        typing.List[typing.Union[str, bool, int]]
    ],
    true: str="yes",
    false: str="no",
    none: str="-",
    delimiter: str=","
) -> str:
    """
    return a string boolean value using parse_bool(), of specified style

    Args:

        true (string):
            The expected return value when data is True

        false (string):
            The expected return value when data is False

        none (string):
            The expected return value when data is None

    Returns:

        string: Map input according to arguments or stringified input

    Usage:

        >>> to_string(True)
        "yes"
        >>> to_string(False)
        "no"

        >>> to_string(True, true="yip", false="nope")
        "yip"
        >>> to_string(False, true="yip", false="nope")
        "nope"

        >>> to_string(None)
        "-"
    """

    if data is None:
        return none

    elif isinstance(data, list) is True:
        children = [
            to_string(x, true, false, none, delimiter)
            for x in list(data)  # type: ignore
            if x is not None
        ]
        if len(children) == 0:
            return none
        normalized_data = ",".join(children)
    else:
        normalized_data = data  # type: ignore

    parsed_data = parse_user_input(normalized_data)

    if parsed_data is True:
        return true
    elif parsed_data is False:
        return false
    elif parsed_data is None:
        return none

    return str(parsed_data)


def exec_passthru(
    command: typing.List[str],
    logger: 'iocage.lib.Logger.Logger'=None
) -> typing.Tuple[str, str]:

    if isinstance(command, str):
        command = [command]

    command_str = " ".join(command)
    if logger:
        logger.spam(f"Executing (interactive): {command_str}")

    return subprocess.Popen(command).communicate()


def exec_raw(
    command: typing.List[str],
    logger: 'iocage.lib.Logger.Logger'=None,
    **kwargs
) -> subprocess.Popen:

    if isinstance(command, str):
        command = [command]

    command_str = " ".join(command)
    if logger:
        logger.spam(f"Executing (raw): {command_str}")

    return subprocess.Popen(
        command,
        **kwargs
    )


def exec_iter(
    command: typing.List[str],
    logger: 'iocage.lib.Logger.Logger'=None
) -> typing.Generator[str, None, None]:

    process = exec_raw(
        command,
        logger=logger,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True
    )

    for stdout_line in iter(process.stdout.readline, ""):
        yield stdout_line

    process.stdout.close()

    return_code = process.wait()
    if return_code:
        raise subprocess.CalledProcessError(return_code, command)


def shell(
    command: typing.Union[str, typing.List[str]],
    logger: 'iocage.lib.Logger.Logger'=None
) -> subprocess.CompletedProcess:

    if not isinstance(command, str):
        shell_command = " ".join(command)
    else:
        shell_command = command

    if logger:
        logger.spam(f"Executing Shell: {command}")

    return subprocess.check_output(
        shell_command,
        shell=True,
        universal_newlines=True,
        stderr=subprocess.DEVNULL
    )


# ToDo: replace with (u)mount library
def umount(
    mountpoint: iocage.lib.Types.AbsolutePath,
    force: bool=False,
    ignore_error: bool=False,
    logger: 'iocage.lib.Logger.Logger'=None
) -> None:

    cmd = ["/sbin/umount"]

    if force is True:
        cmd.append("-f")

    cmd.append(str(mountpoint))

    try:
        exec(cmd)
        if logger is not None:
            logger.debug(
                f"Jail mountpoint {mountpoint} umounted"
            )
    except:
        if logger is not None:
            logger.spam(
                f"Jail mountpoint {mountpoint} not unmounted"
            )
        if ignore_error is False:
            raise iocage.lib.errors.UnmountFailed(
                mountpoint=mountpoint,
                logger=logger
            )


def get_basedir_list(distribution_name="FreeBSD"):
    basedirs = [
        "bin",
        "boot",
        "lib",
        "libexec",
        "rescue",
        "sbin",
        "usr/bin",
        "usr/include",
        "usr/lib",
        "usr/libexec",
        "usr/sbin",
        "usr/share",
        "usr/libdata",
    ]

    if distribution_name == "FreeBSD":
        basedirs.append("usr/lib32")

    return basedirs



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""start module for the cli."""
import click

import iocage.lib.Jails
import iocage.lib.Logger

__rootcmd__ = True


@click.command(name="start", help="Starts the specified jails or ALL.")
@click.pass_context
@click.option("--rc", default=False, is_flag=True,
              help="Will start all jails with boot=on, in the specified"
                   " order with smaller value for priority starting first.")
@click.argument("jails", nargs=-1)
def cli(ctx, rc, jails):
    """
    Starts Jails
    """

    logger = ctx.parent.logger
    start_args = {
        "logger": logger,
        "print_function": ctx.parent.print_events
    }

    if len(jails) == 0:
        logger.error("No jail selector provided")
        exit(1)

    elif rc is True:
        if len(jails) > 0:
            logger.error("Cannot use --rc and jail selectors simultaniously")
            exit(1)
        autostart(**start_args)
    else:
        normal(jails, **start_args)


def autostart(logger, print_function):

    filters = ("boot=yes",)

    ioc_jails = iocage.lib.Jails.JailsGenerator(
        logger=logger,
        filters=filters
    )

    # sort jails by their priority
    jails = sorted(
        list(ioc_jails),
        key=lambda x: x.config["priority"]
    )

    start_jails(jails, logger=logger, print_function=print_function)


def normal(filters, logger, print_function):

    jails = iocage.lib.Jails.JailsGenerator(
        logger=logger,
        filters=filters
    )

    start_jails(jails, logger=logger, print_function=print_function)


def start_jails(jails, logger, print_function):

    changed_jails = []
    failed_jails = []
    for jail in jails:
        try:
            print_function(jail.start())

        except Exception:
            failed_jails.append(jail)
            raise
            continue

        logger.log(f"{jail.humanreadable_name} running as JID {jail.jid}")
        changed_jails.append(jail)

    if len(failed_jails) > 0:
        exit(1)

    if len(changed_jails) == 0:
        jails_input = " ".join(list(jails))
        logger.error(f"No jailes matches your input: {jails_input}")
        exit(1)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import typing
import iocage.lib.errors
import iocage.lib.helpers


class AddressSet(set):

    config: 'iocage.lib.Config.Jail.JailConfig.JailConfig'  # type: ignore

    def __init__(
        self,
        config=None,
        property_name="ip4_address"
    ):

        self.config = config
        set.__init__(self)
        object.__setattr__(self, 'property_name', property_name)

    def add(self, value, notify=True):
        set.add(self, value)
        if notify:
            self.__notify()

    def remove(self, value, notify=True):
        set.remove(self, value)
        if notify:
            self.__notify()

    def __notify(self):
        self.config.update_special_property(self.property_name)


_AddressSetInputType = typing.Union[str, typing.Dict[str, AddressSet]]


class AddressesProp(dict):

    logger: 'iocage.lib.Logger.Logger'
    config: 'iocage.lib.Config.Jail.JailConfig.JailConfig'  # type: ignore
    property_name: str = "ip4_address"
    skip_on_error: bool

    def __init__(
        self,
        config=None,  # type: ignore
        property_name: str="ip4_address",
        logger: 'iocage.lib.Logger.Logger'=None,
        skip_on_error: bool=False
    ) -> None:

        dict.__init__(self, {})

        self.logger = logger
        self.config = config
        self.property_name = property_name
        self.skip_on_error = skip_on_error

    def set(self, data: _AddressSetInputType) -> None:

        self.clear()

        try:
            iocage.lib.helpers.parse_none(data)
            return
        except TypeError:
            pass

        if isinstance(data, dict):
            dict.__init__(self, data)
            return

        ip_addresses = data.split(" ")
        for ip_address_string in ip_addresses:

            try:
                nic, address = ip_address_string.split("|", maxsplit=1)
                self.add(nic, address)
            except ValueError:

                level = "warn" if (self.skip_on_error is True) else "error"

                iocage.lib.errors.InvalidJailConfigAddress(
                    jail=self.config.jail,
                    value=ip_address_string,
                    property_name=self.property_name,
                    logger=self.logger,
                    level=level
                )

                if self.skip_on_error is False:
                    exit(1)

    def add(
        self,
        nic: str,
        addresses: typing.Union[typing.List[str], str]=None,
        notify: bool=True
    ) -> None:

        if addresses is None or addresses == [] or addresses == "":
            return

        if isinstance(addresses, str):
            addresses = [addresses]

        try:
            prop = self[nic]
        except KeyError:
            prop = self.__empty_prop(nic)

        for address in addresses:
            prop.add(address, notify=False)

        if notify:
            self.__notify()

    @property
    def networks(self) -> typing.List[str]:
        """
        Flat list of all networks configured across all nics
        """
        networks: list = []
        for nic, addresses in self.items():
            networks += addresses
        return networks

    def __setitem__(
        self,
        key: str,
        addresses: typing.Union[typing.List[str], str]
    ) -> None:

        try:
            dict.__delitem__(self, key)
        except KeyError:
            pass

        self.add(key, addresses)

    def __delitem__(self, key: str) -> None:
        dict.__delitem__(self, key)
        self.__notify()

    def __notify(self) -> None:
        self.config.update_special_property(self.property_name)

    def __empty_prop(self, key: str) -> AddressSet:
        prop = AddressSet(self.config, property_name=self.property_name)
        dict.__setitem__(self, key, prop)
        return prop

    def __str__(self) -> str:
        if len(self) == 0:
            return ""
        out = []
        for nic in self:
            for address in self[nic]:
                out.append(f"{nic}|{address}")
        return str(" ".join(out))



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import os.path

import ucl

import iocage.lib.helpers
import iocage.lib.Config.Jail.File.Prototype

# MyPy
import iocage.lib.LaunchableResource


class RCConf(
    dict,
    iocage.lib.Config.Jail.File.Prototype.ResourceConfigFile
):

    # the file is always relative to the resource
    _file: str = "/etc/rc.conf"

    def __init__(
        self,
        resource: 'iocage.lib.LaunchableResource.LaunchableResource',
        file: str=None,
        logger: 'iocage.lib.Logger.Logger'=None
    ) -> None:

        dict.__init__(self, {})
        self.logger = iocage.lib.helpers.init_logger(self, logger)

        # No file was loaded yet, so we can't know the delta yet
        self._file_content_changed = True

        if file is not None:
            self._file = file

        self.resource = resource
        self._read_file()

    @property
    def path(self):
        path = f"{self.resource.root_dataset.mountpoint}/{self.file}"
        self._require_path_relative_to_resource(
            filepath=path,
            resource=self.resource
        )
        return os.path.abspath(path)

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, value):
        if self._file != value:
            self._file = value
            self._read_file()

    @property
    def changed(self) -> bool:
        return (self._file_content_changed is True)

    def _read_file(
        self,
        silent: bool=False,
        delete: bool=False,
        merge: bool=False
    ) -> None:
        """
        Read the rc.conf file

        Args:

            silent:
                Do not use the logger

            delete:
                Delete entries that do not exist in the file

            merge:
                Do not change already existing properties
        """
        try:
            if (self.path is not None) and os.path.isfile(self.path):
                data = self._read(silent=silent)
            else:
                data = {}
        except:
            data = {}

        existing_keys = set(self.keys())
        new_keys = set(data.keys())
        delete_keys = existing_keys - new_keys

        if delete is True:
            for key in delete_keys:
                del self[key]

        for key in new_keys:
            if key in existing_keys:
                if (merge is True) and (self[key] != data[key]):
                    self[key] = data[key]
                    self._file_content_changed = True
            else:
                self[key] = data[key]
                self._file_content_changed = True

        if silent is False:
            self.logger.verbose(f"Updated rc.conf data from {self.path}")

        if delete is False and len(delete_keys) > 0:
            # There are properties that are not in the file
            self._file_content_changed = True
        else:
            # Current data matches with file contents
            self._file_content_changed = False

    def _read(self, silent=False):
        data = ucl.load(open(self.path).read())
        self.logger.spam(f"rc.conf was read from {self.path}")
        return data

    def save(self) -> bool:

        if self.changed is False:
            self.logger.debug("rc.conf was not modified - skipping write")
            return False

        with open(self.path, "w") as rcconf:

            output = ucl.dump(self, ucl.UCL_EMIT_CONFIG)
            output = output.replace(" = \"", "=\"")
            output = output.replace("\";\n", "\"\n")

            self.logger.verbose(f"Writing rc.conf to {self.path}")

            rcconf.write(output)
            rcconf.truncate()
            rcconf.close()

            self._file_content_changed = False
            self.logger.spam(output[:-1], indent=1)
            return True

    def __setitem__(self, key, value):
        val = iocage.lib.helpers.to_string(
            iocage.lib.helpers.parse_user_input(value),
            true="YES",
            false="NO"
        )

        try:
            if self[key] == value:
                return
        except KeyError:
            pass

        dict.__setitem__(self, key, val)
        self._file_content_changed = True

    def __getitem__(self, key):
        val = dict.__getitem__(self, key)
        return iocage.lib.helpers.parse_user_input(val)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import iocage.lib.Release
import iocage.lib.Resource
import iocage.lib.helpers

# MyPy
import libzfs


class ReleasesGenerator(iocage.lib.Resource.ListableResource):

    _class_release = iocage.lib.Release.ReleaseGenerator

    def __init__(
        self,
        filters: 'iocage.lib.Filter.Terms'=None,
        host=None,
        zfs=None,
        logger=None
    ) -> None:

        self.logger = iocage.lib.helpers.init_logger(self, logger)
        self.zfs = iocage.lib.helpers.init_zfs(self, zfs)
        self.host = iocage.lib.helpers.init_host(self, host)

        iocage.lib.Resource.ListableResource.__init__(
            self,
            dataset=self.host.datasets.releases,
            filters=filters
        )

    @property
    def local(self):
        release_datasets = self.dataset.children
        return list(map(
            lambda x: self._class_release(
                name=x.name.split("/").pop(),
                logger=self.logger,
                host=self.host,
                zfs=self.zfs
            ),
            release_datasets
        ))

    @property
    def available(self):
        return self.host.distribution.releases

    def _create_resource_instance(
        self,
        dataset: libzfs.ZFSDataset,
        *args,
        **kwargs
    ) -> 'iocage.lib.Release.ReleaseGenerator':

        kwargs["name"] = self._get_asset_name_from_dataset(dataset)
        kwargs["logger"] = self.logger
        kwargs["host"] = self.host
        kwargs["zfs"] = self.zfs
        return self._class_release(*args, **kwargs)

    def _create_instance(self, *args, **kwargs):
        return self._class_release(*args, **kwargs)


class Releases(ReleasesGenerator):

    _class_release = iocage.lib.Release.Release

    def _create_instance(self, *args, **kwargs):
        return iocage.lib.Release.Release(*args, **kwargs)



# Copyright (c) 2014-2017, iocage
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted providing that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import os.path

import iocage.lib.LaunchableResource


class ResourceConfigFile:

    def _require_path_relative_to_resource(
        self,
        filepath: str,
        resource: 'iocage.lib.LaunchableResource.LaunchableResource'
    ) -> None:

        if self._is_path_relative_to_resource(filepath, resource) is False:
            raise iocage.lib.errors.SecurityViolationConfigJailEscape(
                file=filepath
            )

    def _is_path_relative_to_resource(
        self,
        filepath: str,
        resource: 'iocage.lib.LaunchableResource.LaunchableResource'
    ) -> bool:

        real_resource_path = self._resolve_path(resource.dataset.mountpoint)
        real_file_path = self._resolve_path(filepath)

        return real_file_path.startswith(real_resource_path)

    def _resolve_path(self, filepath: str) -> str:
        return os.path.realpath(os.path.abspath(filepath))
